/*
 * Copyright (c) Jupyter Development Team.
 * Distributed under the terms of the Modified BSD License.
 */

/*-----------------------------------------------------------------------------
| Copyright (c) 2014-2018, PhosphorJS Contributors
|
| Distributed under the terms of the BSD 3-Clause License.
|
| The full license is in the file LICENSE, distributed with this software.
|----------------------------------------------------------------------------*/
import '@lumino/dragdrop/style/index';
import '@lumino/widgets/style/index';
import './accordionpanel.css';
import './commandpalette.css';
import './datagrid.css';
import './dockpanel.css';
import './menu.css';
import './menubar.css';
import './scrollbar.css';
import './tabbar.css';
