// Copyright (c) Jupyter Development Team.
// Distributed under the terms of the Modified BSD License.
/*-----------------------------------------------------------------------------
| Copyright (c) 2014-2016, PhosphorJS Contributors
|
| Distributed under the terms of the BSD 3-Clause License.
|
| The full license is in the file LICENSE, distributed with this software.
|----------------------------------------------------------------------------*/
import './array.spec';
import './chain.spec';
import './empty.spec';
import './filter.spec';
import './find.spec';
import './iter.spec';
import './map.spec';
import './range.spec';
import './reduce.spec';
import './repeat.spec';
import './retro.spec';
import './sort.spec';
import './stride.spec';
import './string.spec';
import './take.spec';
import './zip.spec';
