/*
 * Copyright (c) Jupyter Development Team.
 * Distributed under the terms of the Modified BSD License.
 */

/*-----------------------------------------------------------------------------
| Copyright (c) 2014-2018, PhosphorJS Contributors
|
| Distributed under the terms of the BSD 3-Clause License.
|
| The full license is in the file LICENSE, distributed with this software.
|----------------------------------------------------------------------------*/
@import '@lumino/default-theme/style/index.css';

body {
  display: flex;
  flex-direction: column;
  position: absolute;
  top: 0;
  left: 0;
  right: 0;
  bottom: 0;
  margin: 0;
  padding: 0;
  overflow: hidden;
}

#dock {
  flex: 1 1 auto;
  padding: 4px;
}

.content-wrapper {
  padding: 8px;
  border: 1px solid #c0c0c0;
  border-top: none;
  background: white;
  box-shadow: 1px 1px 2px rgba(0, 0, 0, 0.2);
}
