/*
 * Copyright (c) Jupyter Development Team.
 * Distributed under the terms of the Modified BSD License.
 */

/*-----------------------------------------------------------------------------
| Copyright (c) Jupyter Development Team.
| Copyright (c) 2014-2017, PhosphorJS Contributors
|
| Distributed under the terms of the BSD 3-Clause License.
|
| The full license is in the file LICENSE, distributed with this software.
|----------------------------------------------------------------------------*/

.lm-MenuBar {
  padding-left: 5px;
  background: #fafafa;
  color: rgba(0, 0, 0, 0.87);
  border-bottom: 1px solid #dddddd;
  font:
    13px Helvetica,
    Arial,
    sans-serif;
}

.lm-MenuBar-menu {
  transform: translateY(-1px);
}

.lm-MenuBar-item {
  padding: 4px 8px;
  border-left: 1px solid transparent;
  border-right: 1px solid transparent;
}

.lm-MenuBar-item.lm-mod-active {
  background: #e5e5e5;
}

.lm-MenuBar.lm-mod-active .lm-MenuBar-item.lm-mod-active {
  z-index: 10001;
  background: white;
  border-left: 1px solid #c0c0c0;
  border-right: 1px solid #c0c0c0;
  box-shadow: 0px 0px 6px rgba(0, 0, 0, 0.2);
}

.lm-MenuBar-item:focus-visible {
  outline: 2px solid #333;
  outline-offset: -2px;
}

.lm-MenuBar-item[aria-disabled='true'] {
  color: rgba(0, 0, 0, 0.37);
}
