/*
 * Copyright (c) Jupyter Development Team.
 * Distributed under the terms of the Modified BSD License.
 */

/*-----------------------------------------------------------------------------
| Copyright (c) 2014-2018, PhosphorJS Contributors
|
| Distributed under the terms of the BSD 3-Clause License.
|
| The full license is in the file LICENSE, distributed with this software.
|----------------------------------------------------------------------------*/

.lm-DataGrid {
  min-width: 64px;
  min-height: 64px;
  border: 1px solid #a0a0a0;
}

.lm-DataGrid-scrollCorner {
  background-color: #f0f0f0;
}

.lm-DataGrid-scrollCorner::after {
  content: '';
  position: absolute;
  top: 0;
  left: 0;
  width: 1px;
  height: 1px;
  background-color: #a0a0a0;
}

.lm-DataGrid-cellEditorOccluder {
  pointer-events: none;
  position: absolute;
  overflow: hidden;
}

.lm-DataGrid-cellEditorContainer {
  pointer-events: auto;
  position: absolute;
  background-color: #ffffff;
  box-sizing: border-box;
  box-shadow: 0px 0px 6px #006bf7;
  border: 2px solid #006bf7;
}

.lm-DataGrid-cellEditorContainer.lm-mod-invalid {
  box-shadow: 0px 0px 6px red;
  border: 2px solid red;
}

.lm-DataGrid-cellEditorContainer > form {
  width: 100%;
  height: 100%;
  overflow: hidden;
}

.lm-DataGrid-cellEditorWidget {
  width: 100%;
  height: 100%;
  outline: none;
  box-sizing: border-box;
}

.lm-DataGrid-cellEditorInput {
  background-color: #ffffff;
  border: 0;
}

.lm-DataGrid-cellEditorCheckbox {
  margin: 0;
}

.lm-DataGrid-notification {
  position: absolute;
  display: flex;
  overflow: visible;
  animation: fade-in 300ms ease-out;
}

.lm-DataGrid-notificationContainer {
  box-shadow: 0px 2px 5px #999999;
  border-radius: 3px;
  background-color: white;
  color: black;
  border: 1px solid black;
  font-family: sans-serif;
  font-size: 13px;
  padding: 4px;
}

@keyframes fade-in {
  0% {
    opacity: 0;
  }
  50% {
    opacity: 0.7;
  }
  100% {
    opacity: 1;
  }
}
